#include <stdlib.h>
#include "STreeNode.h"

const char* util_TypeToDescription(NodeType t)
{
    static const char* typenames[] =
    {
        "n_id",
        "n_class",
        "n_fdec",
        "n_constrdec",
        "n_methdec",
        "n_paramdec",
        "n_vardec",
        "n_empty",
	"n_expr",
        "n_ret",
        "n_cond",
        "n_while",
        "n_assign",
        "n_relop",
        "n_addop",
        "n_mulop",
        "n_intval",
        "n_doubleval",
        "n_stringval",
        "n_boolval",
        "n_fieldref",
        "n_methcall",
        "n_constrcall",
        "n_cast",
        "n_this",
        "n_new"
    };
    return typenames[t];
}

/* output the required number of spaces to the stream */
void util_pad(FILE* os, unsigned int indent)
{
    int j;
    static const int indentfactor = 2;
    for (j=0; j < indent*indentfactor; ++j)
    {
        fprintf(os," ");
    }
}

STreePtr Create(NodeType t, STreePtr c0, STreePtr c1, STreePtr c2, STreePtr c3)
{
    unsigned int j;
    STreePtr n = (STreePtr)malloc(sizeof(struct STreeNode));
    n->type=t;
    for (j=0; j<MAXCHILDREN; ++j)
    {
        n->children[j]=0;
    }
    n->sibling=0;
    n->children[0]=c0;
    n->children[1]=c1;
    n->children[2]=c2;
    n->children[3]=c3;
    n->extrainfo=0;
    return n;
}

STreePtr CreateString(NodeType t, const char *s)
{
    unsigned int j;
    STreePtr n=(STreePtr)malloc(sizeof(struct STreeNode));
    n->type=t;
    for (j=0; j<MAXCHILDREN; ++j)
    {
        n->children[j]=0;
    }
    n->sibling=0;
    n->extrainfo=0;
    SetExtraInfo(n,s);
    return n;
}


/* Destructor */
void Destroy(STreePtr n)
{
    unsigned int j;
    for (j=0; j<MAXCHILDREN; ++j)
    {
        if (n->children[j] != 0)
        {
            Destroy(n->children[j]);
            n->children[j]=0;
        }
    }
    if (n->sibling!=0)  /* note not a loop */
    {                   /* the rest destroyed recursively */
        Destroy(n->sibling);
        n->sibling=0;
    }
    if (n->extrainfo!=0)
    {
        free(n->extrainfo);
    }
    free(n);
}

STreePtr GetChild(STreePtr n, unsigned int i)
{
    return (i<MAXCHILDREN) ? n->children[i] : 0;
}
        
void SetChild(STreePtr n, unsigned int i, STreePtr c)
{
    if (i<MAXCHILDREN)
    {
        if (n->children[i]!=0)
        {
            Destroy(n->children[i]);
        }
        n->children[i]=c;
    }   
}

bool HasChildren(STreePtr n)
{
    unsigned int j;
    for (j=0; j<MAXCHILDREN; ++j)
    {
        if (n->children[j]!=0)
        {
            return TRUE;
        }
    }
    return FALSE;
}

void SetLastSibling(STreePtr n, STreePtr c)
{
    STreePtr target=n;
    while (target->sibling != 0)
    {
        target=target->sibling;
    }
    SetSibling(target, c);
}


void SetSibling(STreePtr n, STreePtr c)
{
    if (n->sibling!=0)
    {
        Destroy(n->sibling);
    }
    n->sibling=c;
}

void SetExtraInfo(STreePtr n, const char *s)
{
    if (n->extrainfo!=0)
    {
        free(n->extrainfo);
        n->extrainfo=0;         /* in case someone tries to overwrite with null */
    }
    if (s!=0)
    {
        n->extrainfo = (char *)malloc(strlen(s)*sizeof(char)+1);
        strcpy(n->extrainfo, s);
    }
}

void Print(const STreePtr n, FILE* os, unsigned int indent)
{
    unsigned int i, j;
    unsigned int lastchild = 0;
    STreePtr s = n->sibling;

    util_pad(os, indent);
    fprintf(os, "%s %s\n", util_TypeToDescription(n->type), n->extrainfo);
    
    if (HasChildren(n))
    {
        indent++;
        for (j=0; j<MAXCHILDREN; ++j)
        {
            if (n->children[j]!=0)
            {
                lastchild=j;
            }
        }
        for (i=0; i<=lastchild; ++i)
        {
            if (n->children[i]==0)
            {
                util_pad(os, indent);
                fprintf(os, "null\n");
            }
            else
            {
                Print(n->children[i], os, indent);
            }
        }
        indent--;               
    }
    
    // omission of showsibs param needs to be tested...
    if (s!=0)
    {
        Print(s, os, indent);
    }
}

