/**
 * Yacc/bison specification for C implementation of JM parser.
 * J. Fenwick, May 2006.
 * Adapted from C++ to C by R. Topor.
 */

%{
#include <stdio.h>
#include <ctype.h>

#define YYSTYPE STreePtr

#include "STreeNode.h"
#include "ass1.tab.h"

int getToken(void);

YYSTYPE root = 0;
int lnum = 0;

int yylex(void);

void yyerror(const char* msg);

char* yytext;
FILE* yyin;
%}

%token CLASS ELSE ELSEIF EXTENDS IF NEW RETURN THIS VOID WHILE
%token ID BOOLEAN INTEGER DOUBLE STRING
%token ASSIGN EQ NEQ LT LEQ GT GEQ PLUS MINUS TIMES DIVIDES MOD
%token LPAREN RPAREN LBRACK RBRACK LBRACE RBRACE DOT COMMA SEMI
%token NEWLINE		/* Added to make line counting easier */
%token ERROR 

%% /* Grammar for JM */ 

program     : class_seq {$$=$1; root=$1;}
            ;
class_seq   : class_dec {$$=$1;}
            | class_seq class_dec {$$=$1; SetLastSibling($$,$2);}
class_dec   : CLASS ID extension LBRACE declaration_seq RBRACE 
		{$$=Create(n_class,$2,$3,$5,0);}
	    | CLASS ID extension LBRACE RBRACE 
		{$$=Create(n_class,$2,$3,0,0);}
            ;
extension   : EXTENDS ID {$$=$2;}
            | {$$=0;}
            ;
declaration_seq 
            : declaration {$$=$1;}
            | declaration_seq declaration {$$=$1; SetLastSibling($$,$2);}
            ;
declaration : field_dec {$$=$1;}
            | constructor_dec {$$=$1;}
            | method_dec {$$=$1;}
            ;
field_dec   : type_specifier id_seq SEMI {$$=Create(n_fdec,$1,$2,0,0);}
            ;
type_specifier
            : ID {$$=$1;}
            ;
id_seq      : ID {$$=$1;}
            | id_seq COMMA ID {$$=$1; SetLastSibling($1,$3);}
            ;
constructor_dec
            : ID LPAREN parameter_part RPAREN compound_stmt
		{$$=Create(n_constrdec,$1,$3,$5,0);}
            ;

method_dec  : type_specifier ID LPAREN parameter_part RPAREN compound_stmt
		{$$=Create(n_methdec,$1,$2,$4,$6);}
	    | VOID ID LPAREN parameter_part RPAREN compound_stmt
		{$$=Create(n_methdec,0,$2,$4,$6);}
	    ;
parameter_part
            : parameter_dec_seq {$$=$1;}
            | {$$=0;}
            ;
parameter_dec_seq
            : parameter_dec {$$=$1;}
            | parameter_dec_seq COMMA parameter_dec
		{$$=$1; SetLastSibling($$,$3);}
            ;
parameter_dec
            : type_specifier ID {$$=Create(n_paramdec,$1,$2,0,0);}
            ;

statement   : empty_stmt {$$=$1;}
	    | expression_stmt {$$=$1;}
            | return_stmt {$$=$1;}
            | compound_stmt {$$=$1;}
            | conditional_stmt {$$=$1;}
            | iteration_stmt {$$=$1;}
            ;
empty_stmt  : SEMI {$$=Create(n_empty,0,0,0,0);}
            ;
expression_stmt
            : expression SEMI {$$=Create(n_expr,$1,0,0,0);}
            ;
return_stmt : RETURN return_val SEMI {$$=Create(n_ret,$2,0,0,0);}
            ;
return_val  : expression {$$=$1;}
            | {$$=0;}
            ;
compound_stmt
            : LBRACE stmt_component_seq RBRACE
		{$$=$2; if ($$==0) {$$=Create(n_empty,0,0,0,0);} }
            ;
stmt_component_seq
            : stmt_component_seq stmt_component 
		{if ($1==0) {$$=$2;} else {$$=$1; SetLastSibling($$,$2);}}
            |
		{$$=0;}
            ;
stmt_component
            : variable_dec {$$=$1;}
            | statement {$$=$1;}
            ;
variable_dec 
            : type_specifier id_seq SEMI {$$=Create(n_vardec,$1,$2,0,0);}
            ;
conditional_stmt 
            : IF LPAREN expression RPAREN compound_stmt
		{$$=Create(n_cond,$3,$5,0,0);} 
            | IF LPAREN expression RPAREN compound_stmt ELSE compound_stmt
		{$$=Create(n_cond,$3,$5,$7,0);}
	    | IF LPAREN expression RPAREN compound_stmt elseifsection
		{$$=Create(n_cond,$3,$5,$6,0);}
            ;
elseifsection : ELSEIF LPAREN expression RPAREN compound_stmt
		{$$=Create(n_cond,$3,$5,0,0);}
	    | ELSEIF LPAREN expression RPAREN compound_stmt ELSE compound_stmt
		{$$=Create(n_cond,$3,$5,$7,0);}
	    | ELSEIF LPAREN expression RPAREN compound_stmt elseifsection
		{$$=Create(n_cond,$3,$5,$6,0);}
	    ;

iteration_stmt
            : WHILE LPAREN expression RPAREN compound_stmt
		{$$=Create(n_while,$3,$5,0,0);}
            ;
expression  : variable ASSIGN expression {$$=Create(n_assign,$1,$3,0,0);}
            | simple_expression {$$=$1;}
            ;

simple_expression
            : additive_expression relop additive_expression 
		{$$=$2; SetChild($$,0,$1); SetChild($$,1,$3);}
            | additive_expression
		{$$=$1;}
            ;
relop       : EQ {$$=$1;} | NEQ {$$=$1;} | LT {$$=$1;} | LEQ {$$=$1;} | GT {$$=$1;} | GEQ {$$=$1;}
            ;
additive_expression 
            : additive_expression addop term
		{$$=$2; SetChild($$,0,$1); SetChild($$,1,$3);}
            | term
		{$$=$1;}
            ;
addop       : PLUS {$$=$1;} | MINUS {$$=$1;}
            ;
term        : term mulop factor 
		{$$=$2; SetChild($$,0,$1); SetChild($$,1,$3);}
            | factor
		{$$=$1;}
            ;
mulop       : TIMES {$$=$1;} | DIVIDES {$$=$1;} | MOD {$$=$1;}
            ;
factor      : constant {$$=$1;}
            | variable {$$=$1;}
            | constructor_call {$$=$1;}
            | method_call {$$=$1;}
            | LPAREN expression RPAREN {$$=$2;}
            | LBRACK type_specifier RBRACK factor 
		{$$=Create(n_cast,$2,$4,0,0);}
            ;
constant    : BOOLEAN {$$=$1;}
	    | INTEGER {$$=$1;}
	    | DOUBLE {$$=$1;}
	    | STRING {$$=$1;}
	    | THIS {$$=$1;}
            ;
    
variable    : ID {$$=$1;}
            | object DOT ID {$$=Create(n_fieldref,$1,$3,0,0);}
            ;
object      : THIS {$$=$1;}
	    | variable {$$=$1;}
	    | method_call {$$=$1;}
	    | constructor_call {$$=$1;}
	    | LPAREN expression RPAREN {$$=$2;}
	    ;    
constructor_call
            : NEW ID LPAREN argument_part RPAREN
		{$$=Create(n_constrcall,$2,$4,0,0);}
            ;
method_call : ID LPAREN argument_part RPAREN
		{$$=Create(n_methcall,$1,$3,0,0);}
            | object DOT ID LPAREN argument_part RPAREN
		{$$=Create(n_methcall,$1,$5,$3,0);}
            ;
argument_part
            : argument_seq {$$=$1;}
            | {$$=0;}
            ;
argument_seq
            : argument {$$=$1;}
            | argument_seq COMMA argument {$$=$1; SetLastSibling($$,$3);}
            ;
argument    : expression {$$=$1;}
            ;

%%

void yyerror(const char* msg)
{
    fprintf(stderr, "Line %d: %s\n", lnum, msg); 
}

int yylex()
{
    int token;
    while (token=getToken(), token==NEWLINE)
    {
	lnum++;
    }
    const char* t=yytext;
    switch (token)
    {
    case ID : yylval=CreateString(n_id,t); break;
    case EQ : yylval=CreateString(n_relop,"EQ"); break;
    case ASSIGN : yylval=CreateString(n_assign, "ASSIGN"); break;
    case NEQ : yylval=CreateString(n_relop,"NEQ"); break;
    case LT : yylval=CreateString(n_relop,"LT"); break;
    case LEQ : yylval=CreateString(n_relop,"LEQ"); break;
    case GT : yylval=CreateString(n_relop,"GT"); break; 
    case GEQ : yylval=CreateString(n_relop,"GEQ"); break;	
    case PLUS : yylval=CreateString(n_addop,"+"); break;	
    case MINUS : yylval=CreateString(n_addop,"-"); break;	
    case TIMES : yylval=CreateString(n_mulop,"*"); break;	
    case DIVIDES : yylval=CreateString(n_mulop,"/"); break;	
    case MOD : yylval=CreateString(n_mulop,"%"); break;	
    case BOOLEAN : yylval=CreateString(n_boolval,t); break;
    case INTEGER : yylval=CreateString(n_intval,t); break;
    case DOUBLE : yylval=CreateString(n_doubleval,t); break;
    case STRING : yylval=CreateString(n_stringval,t); break;
    case THIS : yylval=CreateString(n_this,"THIS"); break;
    case ERROR : 
	yylval=0;
	printf("Line %d: lexical error at token %s\n", lnum, t); // to stderr
	break;
    default:
	yylval=0;
    }
    return token;
}

int main(int argc, char** argv)
{
    yyin = stdin;
    FILE* listing = fopen("listing", "w");
    if (listing == NULL)
    {
	fprintf(stderr, "Failed to open output file\n");
	return 1;
    }
    if (argc>1)
    {
	yyin = fopen(argv[1], "r");
	if (yyin == NULL)
	{
	    fprintf(stderr, "Failed to open input file.\n");
	    return 1;
	}
    }
    fprintf(stderr, "Creating tree\n");
    root = 0;
    int result=yyparse();
    if (result!=0)
    {
	fprintf(stderr,  "Error in parsing\n");
    }
    else
    {
	fprintf(stderr, "Printing tree\n");
	Print(root, listing, 0);
    }
    fprintf(stderr, "Releasing tree\n");
    Destroy(root); 
    fclose(listing);
    if (yyin!=0)
    {
	fclose(yyin);
    }
    return 0;
}
